import sys

from pyramid.tweens import EXCVIEW
from pyramid.settings import aslist
from pyramid.settings import asbool
from pyramid.httpexceptions import WSGIHTTPException
from pyramid.util import DottedNameResolver

from raven import Client

resolver = DottedNameResolver(None)

PY3 = sys.version_info[0] == 3

if PY3: # pragma: no cover
    import builtins
else:
    import __builtin__ as builtins

def as_globals_list(value):
    L = []
    value = aslist(value)
    for dottedname in value:
        if dottedname in builtins.__dict__:
            if PY3: # pragma: no cover
                dottedname = 'builtins.%s' % dottedname
            else:
                dottedname = '__builtin__.%s' % dottedname
        obj = resolver.resolve(dottedname)
        L.append(obj)
    return L


def crow_tween_factory(handler, registry):

    get = registry.settings.get

    ignored = get('pyramid_crow.ignore', tuple())
    if WSGIHTTPException not in ignored:
        ignored = ignored + (WSGIHTTPException,)

    def crow_tween(request):
        try:
            return handler(request)
        except ignored:
            raise
        except:
            request.raven.captureException()
            raise

    return crow_tween


def _filter_request_body(request):
    """Ensure we don't send an over-long request body to sentry

    65 KB is probably enough for anybody"""
    if request.content_length is None:
        # No content_length - pass through whatever we got.
        return request.body
    elif request.content_length >= 2 ** 16:
        return 'Over-long request body of {} bytes omitted'.format(
            request.content_length
        )
    else:
        return request.body


def _request_to_http_context(request):
    return {
        'method': request.method,
        'url': request.path_url,
        'query_string': request.query_string,
        'data': _filter_request_body(request),
        'headers': dict(request.headers),
        'env': dict(request.environ),
    }


def raven_client(request):
    """Configure a raven client from pyramid config file, add request context

    Uses keys from the pyramid config file beginning with raven. and turns each
    into a kwarg to raven.Client. Also produces http context from request and
    adds to client's context"""
    kwargs = {
        'processors': (
            'pyramid_crow.processors.PyramidSanitizePasswordsProcessor',
        ),
        'timeout': 4,
    }

    for key in request.registry.settings:
        if key.startswith('raven.'):
            kwarg = key.partition('.')[2]
            kwargs[kwarg] = request.registry.settings[key]

    client = Client(**kwargs)
    client.http_context(_request_to_http_context(request))

    return client


def includeme(config):
    get = config.registry.settings.get
    ignored = as_globals_list(get('pyramid_crow.ignore',
                              'pyramid.httpexceptions.WSGIHTTPException'))

    config.registry.settings['pyramid_crow.ignore'] = tuple(ignored)
    config.add_request_method(raven_client, 'raven', reify=True)
    config.add_tween('pyramid_crow.crow_tween_factory', under=EXCVIEW)



from functools import partial

from raven._compat import string_types
from raven.processors import SanitizePasswordsProcessor
from raven.utils import varmap


class PyramidSanitizePasswordsProcessor(SanitizePasswordsProcessor):
    """Do some extra sanitization to pick up places where pyramid tends to
    leak passwords through"""

    def sensitive_repr_filter(self, key, value):
        for field in self.FIELDS:
            if isinstance(value, string_types) and field + '=' in value:
                return '[Filtered]'

        return value

    def filter_stacktrace(self, data):
        """Filter out any local variables that contain sensitive-looking
        patterns in their repr()s"""
        super(PyramidSanitizePasswordsProcessor, self).filter_stacktrace(data)
        for frame in data.get('frames', []):
            if 'vars' not in frame:
                continue

            frame['vars'] = varmap(self.sensitive_repr_filter, frame['vars'])

    def filter_http(self, data):
        """Also descend into env, headers looking for keyval-ish strings"""
        super(PyramidSanitizePasswordsProcessor, self).filter_http(data)
        for n in ('headers', 'env'):
            if isinstance(data[n], dict):
                data[n] = varmap(
                    partial(self.vm_sanitize_keyval, delimiter='&'),
                    data[n],
                )

    def vm_sanitize_keyval(self, key, keyval, delimiter):
        """varmap-friendly way to call _sanitize_keyvals

        Also handles mixed types in env"""
        if isinstance(keyval, string_types):
            return self._sanitize_keyvals(keyval, delimiter)
        else:
            return keyval



import unittest
from pyramid import (
    testing,
    httpexceptions,
)
from raven.utils import json
from webtest import TestApp
import mock

import pyramid_crow


class ExpectedException(Exception):
    pass


class IgnoredException(Exception):
    pass


class TestIntegration(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp(
            settings={
                'raven.dsn': 'https://foo:bar@example.com/notadsn',
            }
        )

    def tearDown(self):
        testing.tearDown()

    def _makeApp(self):
        self.config.include('pyramid_crow')
        app = self.config.make_wsgi_app()
        return TestApp(app)

    def test_noop(self):
        config = self.config

        def view(request):
            return 'ok'

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            resp = app.get('/')

        mock_capture.assert_not_called()
        self.assertEqual(resp.body, b'ok')

    def test_ignored_exception(self):
        config = self.config

        config.registry.settings['pyramid_crow.ignore'] =\
            'pyramid_crow.tests.IgnoredException'

        def view(request):
            raise IgnoredException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            self.assertRaises(IgnoredException, app.get, '/')

        mock_capture.assert_not_called()

    def test_http_exception(self):
        config = self.config

        config.registry.settings['pyramid_crow.ignore'] =\
            'pyramid_crow.tests.IgnoredException'

        def view(request):
            raise httpexceptions.HTTPFound()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            self.assertRaises(httpexceptions.HTTPFound, app.get, '/')

        mock_capture.assert_not_called()

    def test_capture_called(self):
        config = self.config

        def view(request):
            self.request = request
            raise ExpectedException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            self.assertRaises(ExpectedException, app.get, '/')

        mock_capture.assert_called_once()
        self.assertEqual(
            self.request.raven.context['request']['method'], 'GET'
        )
        self.assertEqual(
            self.request.raven.context['request']['url'], 'http://localhost/'
        )
        self.assertEqual(
            self.request.raven.context['request']['data'], ''
        )
        self.assertEqual(
            self.request.raven.context['request']['query_string'], ''
        )
        self.assertEqual(
            self.request.raven.context['request']['headers'],
            {'Host': 'localhost:80'},
        )

    def test_capture_query_string(self):
        config = self.config

        def view(request):
            self.request = request
            raise ExpectedException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            self.assertRaises(ExpectedException, app.get, '/',
                              params=(('foo', 'bar'), ('baz', 'garply')))

        mock_capture.assert_called_once()
        self.assertEqual(
            self.request.raven.context['request']['method'], 'GET'
        )
        self.assertEqual(
            self.request.raven.context['request']['url'], 'http://localhost/'
        )
        self.assertEqual(
            self.request.raven.context['request']['data'], ''
        )
        self.assertEqual(
            self.request.raven.context['request']['query_string'],
            'foo=bar&baz=garply'
        )
        self.assertEqual(
            self.request.raven.context['request']['headers'],
            {'Host': 'localhost:80'},
        )

    def test_capture_body(self):
        config = self.config

        def view(request):
            self.request = request
            raise ExpectedException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'captureException') as mock_capture:
            self.assertRaises(ExpectedException, app.post, '/',
                              params=(('foo', 'bar'), ('baz', 'garply')))

        mock_capture.assert_called_once()
        self.assertEqual(
            self.request.raven.context['request']['method'], 'POST'
        )
        self.assertEqual(
            self.request.raven.context['request']['url'], 'http://localhost/'
        )
        self.assertEqual(
            self.request.raven.context['request']['data'], 'foo=bar&baz=garply'
        )
        self.assertEqual(
            self.request.raven.context['request']['query_string'], ''
        )
        self.assertEqual(
            self.request.raven.context['request']['headers'],
            {
                'Content-Length': '18',
                'Content-Type': 'application/x-www-form-urlencoded',
                'Host': 'localhost:80',
            },
        )

    def test_scrub_sensitive_get(self):
        """Bad form to submit VIA get but let's not leak everywhere"""
        config = self.config

        def view(request):
            self.request = request
            raise ExpectedException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'send') as mock_send:
            self.assertRaises(ExpectedException, app.get, '/',
                              params=(('password', 'ohno'),))

        sentry_data = mock_send.call_args[1]

        dumped_data = json.dumps(sentry_data)
        self.assertNotIn('ohno', dumped_data)

    def test_scrub_sensitive_post(self):
        config = self.config

        def view(request):
            self.request = request
            raise ExpectedException()

        config.add_view(view, name='', renderer='string')

        app = self._makeApp()

        with mock.patch.object(pyramid_crow.Client,
                               'send') as mock_send:
            self.assertRaises(ExpectedException, app.post, '/',
                              params=(('password', 'ohno'),))

        sentry_data = mock_send.call_args[1]

        dumped_data = json.dumps(sentry_data)
        self.assertNotIn('ohno', dumped_data)



from setuptools import setup, find_packages
import sys, os

version = '0.2.1'

setup(name='pyramid_crow',
      version=version,
      description="http context compliant automatic raven integration for pyramid",
      long_description="""\
""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Nicholas Pilon',
      author_email='npilon@gmail.com',
      url='https://github.com/npilon/pyramid_crow',
      license='Apache 2.0',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          'pyramid>=1.2dev',
          'raven>=5.20.0',
      ],
      )
